"""Convert a Keras model to tensorflow protobuf format.
"""
import os
import argparse

import tensorflow as tf
from keras import backend as K

import logging as log

from ergo.project import Project

# @credits to https://stackoverflow.com/questions/45466020/how-to-export-keras-h5-to-tensorflow-pb
def freeze_session(session, keep_var_names=None, output_names=None, clear_devices=True):
    """
    Freezes the state of a session into a pruned computation graph.

    Creates a new computation graph where variable nodes are replaced by
    constants taking their current value in the session. The new graph will be
    pruned so subgraphs that are not necessary to compute the requested
    outputs are removed.
    @param session The TensorFlow session to be frozen.
    @param keep_var_names A list of variable names that should not be frozen,
                          or None to freeze all the variables in the graph.
    @param output_names Names of the relevant graph outputs.
    @param clear_devices Remove the device directives from the graph for better portability.
    @return The frozen graph definition.
    """
    log.info("freezing session and creating pruned computation graph ...")

    graph = session.graph
    with graph.as_default():
        freeze_var_names = list(set(v.op.name for v in tf.global_variables()).difference(keep_var_names or []))
        output_names = output_names or []
        output_names += [v.op.name for v in tf.global_variables()]
        input_graph_def = graph.as_graph_def()
        if clear_devices:
            for node in input_graph_def.node:
                node.device = ""
        frozen_graph = tf.graph_util.convert_variables_to_constants(
            session, input_graph_def, output_names, freeze_var_names)
        return frozen_graph

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo to-tf", description="Convert the model inside an ergo project to TensorFlow format.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("path", help="Path of the project containing the model.")
    args = parser.parse_args(argv)
    return args

def action_to_tf(argc, argv):
    args = parse_args(argv)
    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    elif not prj.is_trained():
        log.error("no trained Keras model found for this projec")
        quit()

    frozen_graph = freeze_session(K.get_session(),
                              output_names=[out.op.name for out in prj.model.outputs])

    log.info("saving protobuf to %s ...", os.path.join(prj.path, 'model.pb'))

    tf.train.write_graph(frozen_graph, prj.path, "model.pb", as_text=False)



import os
import argparse
import logging as log

from ergo.project import Project

def probability(x):
    x = float(x)
    if x < 0 or x > 1:
        raise argparse.ArgumentTypeError("%r not in range [0.0 - 1.0]" % x )
    return x

def validate_args(args):
    if args.dataset is not None and (not args.dataset.startswith('sum://') and not os.path.exists(args.dataset)):
        log.error("file %s does not exist" % args.dataset)
        quit()

    p_train = 1 - args.test - args.validation
    if p_train < 0:
        log.error("validation and test proportion are bigger than 1")
        quit()
    elif p_train < 0.5:
        log.error("using less than 50% of the dataset for training")
        quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo train", description="Start the training phase of a model.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path", help="The path containing the model definition.")

    parser.add_argument( "-d", "--dataset", action="store", dest="dataset", 
        help="Path of the dataset to use or a SUM server address. Leave empty to reuse previously generated subsets.")
    parser.add_argument("-g", "--gpus", action="store", dest="gpus", type=int, default=0, 
        help="Number of GPU devices to use, leave to 0 to use all the available ones.")
    parser.add_argument("-t", "--test", action="store", dest="test", type=probability, default=0.15, 
        help="Proportion of the test set in the (0,1] interval.")
    parser.add_argument("-v", "--validation", action="store", dest="validation", type=probability, default=0.15, 
        help="Proportion of the validation set in the (0,1] interval.")
    
    parser.add_argument("--no-save", dest="no_save", action="store_true", default=False, 
        help="Do not save temporary datasets on disk.")
    parser.add_argument("--no-shuffle", dest="no_shuffle", action="store_true", default=False, 
        help="Do not shuffle dataset during preparation.")

    args = parser.parse_args(argv)
    validate_args(args)
    return args

def action_train(argc, argv):
    args = parse_args(argv)
    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()

    if args.dataset is not None:
        # a dataset was specified, split it and generate
        # the subsets
        prj.dataset.do_save = not args.no_save
        prj.prepare(args.dataset, args.test, args.validation, not args.no_shuffle)
    elif prj.dataset.exists():
        # no dataset passed, attempt to use the previously
        # generated subsets
        prj.dataset.load()
    else:
        log.error("no test/train/validation subsets found in %s, please specify a --dataset argument", args.path)
        quit()

    prj.train(args.gpus)



import os
import sys
import json
import re
import logging as log

import numpy as np
import pandas as pd

from sklearn.metrics import classification_report, confusion_matrix

from keras.models import model_from_yaml, load_model
from keras.utils.training_utils import multi_gpu_model

from ergo.core.utils import serialize_classification_report, serialize_cm
from ergo.core.logic import Logic
from ergo.dataset import Dataset

class Project(object):
    def __init__(self, path):
        # base info
        self.path  = os.path.abspath(path)
        self.logic = Logic(self.path)
        # model related data
        self.model          = None
        self.accu           = None
        self.model_path     = os.path.join(self.path, 'model.yml')
        self.weights_path   = os.path.join(self.path, 'model.h5')
        self.fdeep_path     = os.path.join(self.path, 'model.fdeep')
        # training related data
        self.dataset         = Dataset(self.path)
        self.txt_stats_path  = os.path.join(self.path, 'stats.txt')
        self.json_stats_path = os.path.join(self.path, 'stats.json')
        self.history_path    = os.path.join(self.path, 'history.json')
        self.classes_path    = os.path.join(self.path, 'classes.json')
        self.history         = None
        self.classes         = None
        self.what            = {
            'train' : "Training --------------------------------------------\n",
            'val'   : "Validation ------------------------------------------\n",
            'test'  : "Test ------------------------------------------------\n"
        }

    def exists(self):
        return os.path.exists(self.path)

    def is_trained(self):
        return os.path.exists(self.weights_path)

    def load(self):
        log.info("loading project %s ..." % self.path)

        if not self.exists():
            return "%s does not exist" % self.path

        err = self.logic.load()
        if err is not None:
            return err

        if os.path.exists(self.weights_path):
            log.debug("loading model from %s ...", self.weights_path)
            self.model = load_model(self.weights_path)
            # https://github.com/keras-team/keras/issues/6462
            self.model._make_predict_function()

        elif os.path.exists(self.model_path):
            log.debug("loading model from %s ...", self.model_path)
            with open(self.model_path, 'r') as fp:
                self.model = model_from_yaml(fp.read())

        else:
            self.model = self.logic.builder(True)

        if os.path.exists(self.history_path):
            log.debug("loading history from %s ...", self.history_path)
            with open(self.history_path, 'r') as fp:
                self.history = json.load(fp)


        if os.path.exists(self.classes_path):
            log.debug("loading classes from %s ...", self.classes_path)
            with open(self.classes_path, 'r') as fp:
                self.classes = {int(k) : v for k, v in json.load(fp).items()}

        return None

    def accuracy_for(self, X, Y, repo_as_dict = False):
        Y_tpred = np.argmax(self.model.predict(X), axis = 1)
        repo    = classification_report(np.argmax(Y, axis = 1), Y_tpred, output_dict = repo_as_dict)
        cm      = confusion_matrix(np.argmax(Y, axis = 1), Y_tpred)
        return repo, cm

    def accuracy(self):
        train, tr_cm = self.accuracy_for(self.dataset.X_train, self.dataset.Y_train)
        test,  ts_cm = self.accuracy_for(self.dataset.X_test, self.dataset.Y_test)
        val,  val_cm = self.accuracy_for(self.dataset.X_val, self.dataset.Y_val)
        return {'train': (train, tr_cm),
                'test': (test, ts_cm),
                'val': (val, val_cm)}

    def _save_model(self):
        log.info("updating %s ...", self.model_path)
        with open( self.model_path, 'w' ) as fp:
            fp.write(self.model.to_yaml())

        log.info("updating %s ...", self.weights_path)
        self.model.save(self.weights_path)

    def _save_history(self):
        log.info("updating %s ...", self.history_path)
        with open(self.history_path, 'w') as fp:
            json.dump(self.history, fp)

    def _emit_txt_stats(self, where):
        for who, header in self.what.items():
            vals = self.accu[who]
            where.write( header )
            where.write( vals[0] )
            where.write("\n\n")
            where.write("confusion matrix:")
            where.write("\n\n")
            where.write("%s\n" % vals[1])
            where.write("\n")

    def _emit_json_stats(self, where):
        stats = {}
        for who in self.what:
            report, cm = self.accu[who]
            stats[who] = { 
                'accuracy': serialize_classification_report(report), 
                'cm': serialize_cm(cm)
            }
        json.dump(stats, where)

    def _save_stats(self):
        log.info("updating %s ...", self.txt_stats_path)
        with open(self.txt_stats_path, 'w') as fp:
            self._emit_txt_stats(fp)

        log.info("updating %s ...", self.json_stats_path)
        with open(self.json_stats_path, 'wt') as fp:
            self._emit_json_stats(fp)

    def _from_file(self, filename):
        log.info("preparing data from %s ...", filename)
        return self.logic.prepare_dataset(filename)

    def prepare(self, source, p_test, p_val, shuffle = True):
        data = self._from_file(source)
        num_labels = None
        if self.model is not None:
            # assuming only one single dense output layer
            num_labels = self.model.outputs[-1].shape[1]
        log.info("data shape: %s", data.shape)
        return self.dataset.source(data, p_test, p_val, shuffle, num_labels)

    def train(self, gpus):
        # async datasets saver might be running, wait before training
        self.dataset.saver.wait()

        # train
        if self.model is None:
            self.model = self.logic.builder(True)

        to_train = self.model
        if gpus > 1:
            log.info("training with %d GPUs", gpus)
            to_train = multi_gpu_model(self.model, gpus=gpus)

        past = self.history.copy() if self.history is not None else None
        present = self.logic.trainer(to_train, self.dataset).history

        if past is None:
            self.history = present
        else:
            self.history = {}
            for name, past_values in past.items():
                self.history[name] = past_values + present[name]

        self.accu = self.accuracy()

        print("")
        self._emit_txt_stats(sys.stdout)

        # save model structure and weights
        self._save_model()
        # save training history
        self._save_history()
        # save model accuracy statistics
        self._save_stats()
          
    def view(self, img_only = False):
        import ergo.views as views

        views.model(self, img_only)
        views.roc(self, img_only)
        views.stats(self, img_only)
        views.history(self, img_only)
        views.show(img_only)



import argparse
import json
import logging as log
import time
import numpy as np

from terminaltables import AsciiTable

from ergo.core.queue import TaskQueue
from ergo.project import Project

# https://stackoverflow.com/questions/11942364/typeerror-integer-is-not-json-serializable-when-serializing-json-in-python
def default(o):
    if isinstance(o, np.int64): return int(o)
    raise TypeError

def validate_args(args):
    if args.ratio > 1.0 or args.ratio <= 0:
        log.error("ratio must be in the (0.0, 1.0] interval")
        quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo relevance", description="Compute the relevance of each feature of the dataset by differential evaluation.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path", help="Path of the project.")

    parser.add_argument("-d", "--dataset", dest="dataset", action="store", type=str, required=True,
        help="Dataset file to use.")
    parser.add_argument("-m", "--metric", dest="metric", action="store", type=str, default='precision',
        help="Which metric to track changes of while performing differential evaluation.")
    parser.add_argument("-a", "--attributes", dest="attributes", action="store", type=str, required=False,
        help="Optional file containing the attribute names, one per line.")
    parser.add_argument("-r", "--ratio", dest="ratio", action="store", type=float, required=False, default=1.0,
        help="Size of the subset of the dataset to use in the (0,1] interval.")
    parser.add_argument("-j", "--to-json", dest="to_json", action="store", type=str, required=False,
        help="Output the relevances to this json file.")
    parser.add_argument("-w", "--workers", dest="workers", action="store", type=int, default=0,
        help="If 0, the algorithm will run a single thread iteratively, otherwise create a number of workers " + \
             "to distribute the computation among, -1 to use the number of logical CPU cores available. WARNING: " + \
             "The dataset will be copied in memory for each worker."   )

    args = parser.parse_args(argv)
    return args

def get_attributes(filename, ncols):
    attributes = []
    if filename is not None:
        with open(filename) as f:
            attributes = f.readlines()
        attributes = [name.strip() for name in attributes]
    else:
        attributes = ["feature %d" % i for i in range(0, ncols)]

    return attributes

def zeroize_feature(X, col, is_scalar_input):
    if is_scalar_input:
        backup = X[:, col].copy()
        X[:,col] = 0
    else:
        backup = X[col].copy()
        X[col] = np.zeros_like(backup)

    return backup

def restore_feature(X, col, backup, is_scalar_input):
    if is_scalar_input:
        X[:,col] = backup
    else:
        X[col] = backup


prj    = None
deltas = []
tot    = 0
ncols  = 0
nrows  = 0
start  = None
speed  = 0
attributes = None

def run_inference_without(X, y, col, flat, ref, metric):
    global prj, deltas, tot, start, speed, nrows, ncols, attributes

    log.info("[%.2f evals/s] computing relevance for attribute [%d/%d] %s ...", speed, col + 1, ncols, attributes[col])

    copy = X.copy()
    zeroize_feature(copy, col, flat)

    start = time.time()

    accu, cm = prj.accuracy_for(copy, y, repo_as_dict = True)

    speed = (1.0 / (time.time() - start)) * nrows

    delta = ref - accu['weighted avg'][metric]
    tot  += delta

    deltas.append((col, delta))

def action_relevance(argc, argv):
    global prj, deltas, tot, start, speed, nrows, ncols, attributes

    args = parse_args(argv)
    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    elif not prj.is_trained():
        log.error("no trained Keras model found for this project")
        quit()

    prj.prepare(args.dataset, 0.0, 0.0)

    # one single worker in blocking mode = serial
    if args.workers == 0:
        args.workers = 1

    X, y         = prj.dataset.subsample(args.ratio)
    nrows, ncols = X.shape if prj.dataset.is_flat else (X[0].shape[0], len(X))
    attributes   = get_attributes(args.attributes, ncols)
    queue        = TaskQueue('relevance', num_workers=args.workers, blocking=True)
   
    if args.workers == 1:
        log.info("computing relevance of %d attributes on %d samples using '%s' metric (slow mode) ...", ncols, nrows, args.metric)
    else:
        log.info("computing relevance of %d attributes on %d samples using '%s' metric (parallel with %d workers) ...", ncols, nrows, args.metric, queue.num_workers)

    start = time.time()
    ref_accu, ref_cm = prj.accuracy_for(X, y, repo_as_dict = True)
    speed = (1.0 / (time.time() - start)) * nrows

    for col in range(0, ncols):
        queue.add_task( run_inference_without, 
                X, y, col, prj.dataset.is_flat, 
                ref_accu['weighted avg'][args.metric], 
                args.metric) 

    # wait for all inferences to finish
    queue.join()
    # sort relevances by absolute value
    deltas = sorted(deltas, key = lambda x: abs(x[1]), reverse = True)

    rels     = []
    num_zero = 0
    table    = [("Column", "Feature", "Relevance")]

    for delta in deltas:
        col, d = delta
        colname = attributes[col]
        rel = {
            "attribute": colname,
            "index": col, 
            "relevance": 0.0     
        }

        if d != 0.0:
            relevance = (d / tot) * 100.0
            row       = ("%d" % col, attributes[col], "%.2f%%" % relevance)
            row       = ["\033[31m%s\033[0m" % e for e in row] if relevance < 0.0 else row
            table.append(row)
            rel['relevance'] = relevance
        else:
            num_zero += 1

        rels.append(rel)

    print("")
    print(AsciiTable(table).table)
    print("")

    if num_zero > 0:
        log.info("%d features have 0 relevance.", num_zero)

    if args.to_json is not None:
        print("")
        log.info("creating %s ...", args.to_json)
        with open(args.to_json, 'w+') as fp:
            json.dump(rels, fp, default=default)



from ergo.core.template import Template

prepare = \
"""
import pandas as pd

# this function is called whenever the `ergo train <project> --dataset file.csv`
# command is executed, the first argument is the dataset and it must return
# a pandas.DataFrame object.
def prepare_dataset(filename):
    # simply read as csv
    return pd.read_csv(filename, sep = ',', header = None)

# this function is called to process a single input into a vector
# that can be used for training or to run an inference.
# it is called from `ergo encode ...` with `is_encoding` set to True,
# in which case you can add additional metadata to the vector, or
# from the `ergo serve ...` API, in which case the vector will be
# used for inference and can't contain metadata.
def prepare_input(x, is_encoding = False):
    # simply read as csv
    return pd.read_csv( pd.compat.StringIO(x), sep = ',', header = None)
"""

model = \
"""
import logging as log

from keras.models import Sequential
from keras.layers import Dense, Activation, Dropout

# build the model
def build_model(is_train):  
    n_inputs       = {NUM_INPUTS}
    n_hidden       = [{HIDDEN}]
    dropout        = 0.4
    activation     = 'relu'
    out_activation = 'softmax'
  
    log.info("building model for %s ..." % 'training' if is_train else 'evaluation')

    model = Sequential()
    for i, n_neurons in enumerate(n_hidden):
        # setup the input layer
        if i == 0:
            model.add(Dense(n_neurons, input_shape = (n_inputs,), activation = activation))
        else:
            model.add(Dense(n_neurons, activation = activation))
        # add dropout
        if is_train:
            model.add(Dropout(dropout))
    # setup output layer
    model.add(Dense({NUM_OUTPUTS}, activation = out_activation))
    
    return model
"""

train = \
"""
import logging as log

from keras.callbacks import EarlyStopping

# define training strategy
def train_model(model, dataset):
    log.info("training model (train on %d samples, validate on %d) ..." % ( \\
            len(dataset.Y_train), 
            len(dataset.Y_val) ) )
    
    loss      = 'categorical_crossentropy'
    optimizer = 'adam'
    metrics   = ['accuracy']
    
    model.compile(loss = loss, optimizer = optimizer, metrics = metrics)

    earlyStop = EarlyStopping(monitor = 'val_acc', min_delta=0.0001, patience = 5, mode = 'auto')
    return model.fit( dataset.X_train, dataset.Y_train,
            batch_size = {BATCH_SIZE},
            epochs = {MAX_EPOCHS},
            verbose = 2,
            validation_data = (dataset.X_val, dataset.Y_val),
            callbacks = [earlyStop])
"""

deps = \
"""
ergo-ai
"""

gitignore = \
"""
*.pyc
*.csv
__pycache__
"""

Templates = [ 
    Template('prepare.py', prepare),
    Template('model.py', model),
    Template('train.py', train),
    Template('requirements.txt', deps),
    Template('.gitignore', gitignore)
]



import argparse
from terminaltables import AsciiTable
import logging as log
import numpy as np
import pandas as pd
np.seterr(divide='ignore', invalid='ignore')

from ergo.project import Project
import ergo.views as views

def validate_ratio(args):
    if args.ratio > 1.0 or args.ratio <= 0:
        log.error("ratio must be in the (0.0, 1.0] interval")
        quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo explore", description="Explore dataset properties",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("path", help="Path of the project to explore.")
    parser.add_argument("-a", "--attributes", dest="attributes", action="store", type=str, required=False,
                        help="Optional file containing the attribute names, one per line.")
    parser.add_argument("-r", "--ratio", dest="ratio", action="store", type=validate_ratio, required=False, default=1.0,
                        help="Size of the subset of the dataset to use in the (0,1] interval.")
    parser.add_argument("-d", "--dataset", dest="dataset", action="store", type=str, required=True,
                        help="Dataset file to use.")
    parser.add_argument("--img-only", dest="img_only", default=False, action="store_true",
                        help="Save plots as PNG files but don't show them in a UI.")
    parser.add_argument("-s", dest="stats", action="store_true", default=False,
                        help="Show dataset statistics")
    parser.add_argument("-c", dest="correlations", action="store_true", default=False,
                        help="Show correlation tables and graphs")
    parser.add_argument("-p", dest="pca", action="store_true", default=False,
                        help="Calculate pca decomposition and explained variance")
    parser.add_argument("-k", dest="cluster", action="store_true", default=False,
                        help="Perform clustering analysis")
    parser.add_argument("--algorithm", dest="cluster_alg", choices=['kmeans', 'dbscan'], default='kmeans', const='kmeans', nargs='?',
                        help="Algorithm for clustering analysis")
    parser.add_argument("-n", dest="nclusters", action="store", required = False, type=float,
                        help="Number of clusters for Kmeans algorithm")
    parser.add_argument("--nmax", dest="nmaxclusters", action="store", required=False, type=int,
                        help="Perform inertia analysis using nmax clusters at most")
    parser.add_argument("--3D", dest="D3", action="store_true", default=False,
                        help="Plot 3D projections of the data")
    parser.add_argument("--all", dest="all", action="store_true", default=False,
                        help="Process all capabilities of ergo explore (can be time consuming deppending on dataset")
    parser.add_argument("-w", "--workers", dest="workers", action="store", type=int, default=0,
                        help="If 0, all the algorithms will run a single thread. Otherwise all algorithms will use w number of jobs "+ \
                             "to distribute the computation among, -1 to use the number of logical CPU cores available.")
    return parser.parse_args(argv)

def red(s):
    return "\033[31m" + s + "\033[0m"

def terminal(s):
    return s

def get_attributes(filename, ncols):
    attributes = []
    if filename is not None:
        with open(filename) as f:
            attributes = f.readlines()
        attributes = [name.strip() for name in attributes]
    else:
        attributes = ["feature %d" % i for i in range(0, ncols)]

    return attributes

prj = None
ncols = 0
nrows = 0
attributes = None
n_jobs = 1


def compute_correlations_with_target (X,y):
    global nrows, ncols, attributes
    y = np.argmax(y, axis=1)
    corr = [(attributes[i], np.corrcoef(X[:, i], y)[0,1], i)
            for i in range(0, ncols)]
    corr = [(i, j, k) for i, j, k in corr if not np.isnan(j)]
    corr = sorted(corr, key=lambda x: abs(x[1]), reverse=True)
    return corr


def print_target_correlation_table(corr, min_corr = 0.3):
    table = [("Column", "Feature", "Correlation with target")]
    not_relevant = 0
    for at, c, idx in corr:
        if abs(c) < min_corr:
            not_relevant += 1
            continue
        row = ( "%d" % idx, "%s" % at, "%.4f" % c)
        table.append(row)

    print("")
    print(AsciiTable(table).table)
    print("")

    if not_relevant > 0:
        log.info("%d features have correlation lower than %f with target attribute.", not_relevant, min_corr)

def calculate_pca(X):
    from sklearn.decomposition import PCA
    dec = PCA()
    dec.fit(X)
    return dec


def calculate_corr(X):
    global attributes
    return pd.DataFrame(np.corrcoef(X, rowvar=False), columns=attributes, index=attributes)

def is_in_table(table, entry):
    for row in table:
        if row[1] == entry[0] and row[0] == entry[1]:
            return True
    return False


def print_correlation_table(corr, min_corr=0.6):
    abs_corr = corr.abs().unstack()
    sorted_corr = abs_corr.sort_values(kind = "quicksort", ascending=False)
    table = [("Feature", "Feature", "Correlation")]
    for idx in sorted_corr.index:
        if idx[0] == idx[1]:
            continue
        if sorted_corr.loc[idx] <= min_corr:
            break
        if is_in_table(table, idx):
            continue
        table.append(( idx[0], idx[1], "%.3f" % sorted_corr.loc[idx]))
    print("")
    log.info("Showing variables with a correlation higher than %f" % min_corr)
    print("")
    print(AsciiTable(table).table)
    print("")


def print_stats_table(X):
    global attributes
    minv = X.min(axis = 0)
    maxv = X.max(axis = 0)
    stdv = X.std(axis = 0)

    table = [("Feature", "Min value", "Max value", "Standard deviation")]
    constant = []
    unormalized = []
    for a, mi, ma, st in zip(attributes, minv, maxv, stdv ):
        color = terminal
        if mi == ma:
            constant.append(a)
            continue
        if mi < -1 or ma > 1:
            color = red
            unormalized.append(color(a))

        table.append( (color(a), color("%.2e" % mi) , color("%.2e" % ma), color( "%.2e"% st)))

    print("")
    log.info("Features distribution:")
    print("")
    print(AsciiTable(table).table)
    print("")
    log.warning("The following attributes have constant values: %s" % ', '.join(constant))
    print("")
    log.warning("The following features are not normalized: %s" % ', '.join(unormalized))


def kmeans_clustering(X, n_clusters):
    global n_jobs
    from sklearn.cluster import KMeans
    km = KMeans(n_clusters = n_clusters, n_jobs=n_jobs)
    km.fit(X)
    return km


def dbscan_clustering(X, var):
    global n_jobs
    from sklearn.cluster import DBSCAN
    db = DBSCAN(var, min_samples=10, n_jobs=n_jobs)
    db.fit(X)
    return db


def n_clusters_analysis(X, nmax = 10, nmin = 2):
    global prj, n_jobs
    from sklearn.cluster import KMeans

    inertia = []
    for n in range(nmin, nmax + 1):
        log.info("fitting %d clusters to data" % n)
        km = KMeans(n_clusters = n, n_jobs=n_jobs).fit(X)
        inertia.append(km.inertia_)
    views.plot_intertia(prj, range(nmin, nmax + 1), inertia)


def action_explore(argc, argv):
    global prj, nrows, ncols, attributes, n_jobs

    args     = parse_args(argv)

    if args.all:
        args.pca = True
        args.correlations = True
        args.stats = True
        args.cluster = True
        args.D3 = True

    if args.workers == -1:
        import multiprocessing
        n_jobs = multiprocessing.cpu_count()
    elif args.workers != 0:
        n_jobs = args.workers
    log.info("using %d workers" % n_jobs)

    if args.nclusters and not args.cluster:
        log.warning("number of clusters specified but clustering won't be perfomed")

    if not (args.pca or args.correlations or args.stats or args.cluster):
        log.error("No exploration action was specified")
        print("")
        parse_args(["-h"])
        quit()

    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()

    prj.prepare(args.dataset, 0.0, 0.0)
    if not prj.dataset.is_flat:
        log.error("data exploration can only be applied to flat inputs")
        quit()

    X, y = prj.dataset.subsample(args.ratio)
    nrows, ncols = X.shape
    attributes = get_attributes(args.attributes, ncols)

    if args.correlations:
        log.info("computing correlations of each feature with target")
        corr = compute_correlations_with_target(X,y)
        print_target_correlation_table(corr)
        log.info("computing features crosscorrelation")
        corr = calculate_corr(X)
        print_correlation_table(corr, min_corr=0.7)
        views.correlation_matrix(prj, corr, args.img_only)

    if args.pca:
        log.info("computing pca")
        pca = calculate_pca(X)
        log.info("computing pca projection")
        views.pca_projection(prj, pca, X, y, False)
        if args.D3:
            views.pca_projection(prj, pca, X, y, args.D3)
        views.pca_explained_variance(prj, pca, args.img_only)

    if args.stats:
        log.info("computing features stats")
        print_stats_table(X)

    inertia = False
    if args.cluster:
        if args.cluster_alg == 'kmeans':
            cluster_alg = kmeans_clustering
            if not args.nclusters:
                args.nclusters = len(set(np.argmax(y, axis=1)))
            args.nclusters = int(args.nclusters)
            if args.nmaxclusters:
                log.info("performing inertia analysis with clusters in the range (%d, %d)" % (args.nclusters, args.nmaxclusters))
                inertia = True
                n_clusters_analysis(X, args.nmaxclusters, args.nclusters)
            else:
                log.info("computing kmeans clustering with k=%d" % args.nclusters)
        elif args.cluster_alg == 'dbscan':
            cluster_alg = dbscan_clustering
            if not args.nclusters:
                args.nclusters = 2
            log.info("computing dbscan clustering with eps=%f" % args.nclusters)
            if args.nmaxclusters:
                log.warning("nmax specified but not used. Inertia analysis only available for Kmeans.")
        if not args.pca and not inertia:
            log.info("computing pca to plot clusters")
            pca = calculate_pca(X)
        if not inertia:
            ca = cluster_alg(X, args.nclusters)
            if len(set(ca.labels_)) == 1:
                log.error("clustering failed. Check input parameter.")
                quit()
            views.plot_clusters(prj, pca, X, y, ca, False)
            if args.D3:
                views.plot_clusters(prj, pca, X, y, ca, args.D3)

    views.show(args.img_only)



import os
import json
import argparse
import logging as log
import numpy as np
import tempfile

from terminaltables import AsciiTable

from ergo.project import Project


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo cmp",
                                     description="Compare the performances of two models against a given dataset.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path_1", help="Path of the first project to compare.")
    parser.add_argument("path_2", help="Path of the second project to compare.")

    parser.add_argument("-d", "--dataset", dest="dataset", action="store", type=str, required=True,
                        help="The dataset file to use to compare the models.")
    parser.add_argument("-j", "--to-json", dest="to_json", action="store", type=str, required=False,
                        help="Output the comparision results to this json file.")

    args = parser.parse_args(argv)
    return args


def red(s):
    return "\033[31m" + s + "\033[0m"


def green(s):
    return "\033[32m" + s + "\033[0m"


# https://stackoverflow.com/questions/11942364/typeerror-integer-is-not-json-serializable-when-serializing-json-in-python
def default(o):
    if isinstance(o, np.int64): return int(o)
    raise TypeError


def generate_reduced_dataset(dataset, size=10):
    temp_name = next(tempfile._get_candidate_names())
    tmpfile = open(temp_name, 'w')
    count = 0
    with open(dataset, 'r') as inpfile:
        for line in inpfile: count += 1

    size = min(count, 100)
    size = max(count // 100, size)
    log.info("creating temporary dataset %s of size %d", temp_name, size)

    with open(dataset, 'r') as inpfile:
        p = float(size) / count
        # assuming header:
        tmpfile.write(inpfile.readline())
        for i in inpfile:
            line = inpfile.readline()
            if np.random.random() < p:
                tmpfile.write(line)
    tmpfile.close()
    return temp_name


def are_preparation_equal(projects, dataset):
    inp_shape = None
    for prj in projects:
        if inp_shape is None:
            inp_shape = prj.model.input_shape
        elif inp_shape != prj.model.input_shape:
            return False
        prj.prepare(dataset, 0.0, 0.0, False)
    return compare_datasets(projects[0].dataset, projects[1].dataset)


def compare_datasets(ds1, ds2):
    if ds1.is_flat:
        return np.array_equal(ds1.X, ds2.X)
    for i, j in zip(ds1.X, ds2.X):
        if not np.array_equal(i, j):
            return False
    return True


def action_compare(argc, argv):
    args = parse_args(argv)
    metrics = {}
    projects = { \
        args.path_1: None,
        args.path_2: None,
    }
    ref = None
    inp_shape = None
    out_shape = None
    is_prepared = None
    prjs = []

    for path in projects:
        prj = Project(path)
        err = prj.load()
        if err is not None:
            log.error("error while loading project %s: %s", path, err)
            quit()
        prjs.append(prj)
        if not is_prepared:
            is_prepared = True
        else:
            small_dataset = generate_reduced_dataset(args.dataset)
            are_equal = are_preparation_equal(prjs, small_dataset)
            log.info("deleting temporal file %s", small_dataset)
            os.remove(small_dataset)
        if out_shape is None:
            out_shape = prj.model.output_shape
        elif out_shape != prj.model.output_shape:
            log.error("model %s output shape is %s, expected %s", path, prj.model.output_shape, out_shape)
            quit()
        projects[path] = prj

    for prj, path in zip(prjs, projects):
        prj = Project(path)
        err = prj.load()

        if err is not None:
            log.error("error while loading project %s: %s", path, err)
            quit()

        if ref is None:
            prj.prepare(args.dataset, 0, 0, False)
            ref = prj
            is_prepared = True
        else:
            if are_equal:
                log.info("Projects use same prepare.py file ...")
                prj.dataset.X, prj.dataset.Y, prj.dataset.n_labels = ref.dataset.X.copy(), ref.dataset.Y.copy(), ref.dataset.n_labels
            else:
                log.info("Projects use different prepare.py files, reloading dataset ...")
                prj.prepare(args.dataset, 0., 0., False)

        # TODO: Run in parallel?
        log.debug("running %s ...", path)
        metrics[path] = prj.accuracy_for(prj.dataset.X, prj.dataset.Y, repo_as_dict=True)

    prev = None
    for path, m in metrics.items():
        if prev is None:
            prev = m
            continue

        ref_repo, ref_cm = prev
        new_repo, new_cm = m
        diffs = {
            'report': [],
            'cm': [],
            'cm_stats': {}
        }

        table = [["Name", "Ref", "New", "Delta"]]
        for label, ref_run in ref_repo.items():
            for name, ref_value in ref_run.items():
                new_value = new_repo[label][name]
                if new_value != ref_value:
                    delta = new_value - ref_value
                    sign, fn = ('+', green) if delta >= 0 else ('', red)
                    diffs['report'].append({
                        'name': '%s / %s' % (label, name),
                        'delta': delta,
                    })
                    table.append([ \
                        "%s / %s" % (label, name),
                        "%.2f" % ref_value,
                        "%.2f" % new_value,
                        fn("%s%.2f" % (sign, delta))])
        print("")
        print(AsciiTable(table).table)

        heads = [""]
        for i in range(0, ref_cm.shape[0]):
            heads.append("class %d" % i)

        table = [heads]
        total = 0
        impr = 0
        regr = 0
        for i in range(0, ref_cm.shape[0]):
            row = ["class %d" % i]
            row_diffs = []
            for j in range(0, ref_cm.shape[1]):
                ref_v = ref_cm[i][j]
                new_v = new_cm[i][j]
                total = total + new_v
                delta = new_v - ref_v

                if ref_v != new_v:
                    sign = '+' if delta >= 0 else ''

                    if i == j:
                        fn = green if delta >= 0 else red
                    else:
                        fn = red if delta >= 0 else green

                    if fn == green:
                        impr += abs(delta)
                    else:
                        regr += abs(delta)

                    cell = fn("%d (%s%d)" % (new_v, sign, delta))
                else:
                    cell = "%d" % ref_v

                row.append(cell)
                row_diffs.append(delta)

            diffs['cm'].append(row_diffs)
            table.append(row)

        print("")
        print(AsciiTable(table).table)

        diffs['cm_stats'] = {
            'improvements': {
                'total': impr,
                'perc': impr / float(total) * 100.0
            },
            'regressions': {
                'total': regr,
                'perc': regr / float(total) * 100.0
            }
        }

        print("")
        print("Improvements: %d ( %.2f %% )" % (impr, impr / float(total) * 100.0))
        print("Regressions : %d ( %.2f %% )" % (regr, regr / float(total) * 100.0))

        if args.to_json is not None:
            print("")
            log.info("creating %s ...", args.to_json)
            with open(args.to_json, 'w+') as fp:
                json.dump(diffs, fp, default=default)



import os
import argparse

from ergo.core.utils import clean_if_exist

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo clean", description="Clean a project from temporary datasets and optionally reset it to its initial state.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path", help="Path of the project to clean.")
    parser.add_argument( "-a", "--all", dest="all", action="store_true", default=False,
        help="Remove model weights and training data.")

    args = parser.parse_args(argv)
    return args

def action_clean(argc, argv):
    args = parse_args(argv)
    files = [ \
        'data-train.csv',
        'data-test.csv',
        'data-validation.csv',
        'data-train.pkl',
        'data-test.pkl',
        'data-validation.pkl']

    if args.all:
        files += [ \
            '__pycache__',
            'logs',
            'model.yml',
            'model.h5',
            'model.fdeep',
            'model.stats', # legacy
            'model.png', # legacy
            'test_cm.png',
            'training_cm.png',
            'validation_cm.png',
            'history.png',
            'roc.png',
            'stats.txt',
            'stats.json',
            'history.json',
            'pca_projection.png',
            'pca_explained_ratio.png',
            'corr_matrix.png'
        ]
    
    clean_if_exist(args.path, files)



import os
import argparse
import traceback
import logging as log
import pandas as pd
import numpy as np
from flask import Flask, request, jsonify

from ergo.project import Project

prj         = None
classes     = None
num_outputs = 0
app         = Flask(__name__)

def get_input(req):
    try:
        # search in query parameters
        x = request.args.get('x')
        if x is not None and x != "":
            return x
    except:
        pass

    try:
        # search in form body
        x = request.form.get('x')
        if x is not None and x != "":
            return x
    except:
        pass


    try:
        # search as file upload
        x = request.files['x']
        if x is not None:
            return x
    except:
        pass

    try:
        # search for direct json body
        x = req.get_json(silent=True)
        if x is not None:
            return x
    except:
        pass

    return None

@app.route('/encode', methods=['POST', 'GET'])
def encode_route():
    global prj, classes, num_outputs

    try:
        xin = get_input(request)
        if xin is None:
            return "missing 'x' parameter", 400

        x = prj.logic.prepare_input(xin)

        return jsonify(list(x)), 200
    except Exception as e:
        return str(e), 400

@app.route('/', methods=['POST', 'GET'])
def infer_route():
    global prj, classes, num_outputs
    try:
        xin = get_input(request)
        if xin is None:
            return "missing 'x' parameter", 400

        # encode the input
        x = prj.logic.prepare_input(xin)
        # run inference
        y = prj.model.predict(np.array([x]))[0].tolist()
        # decode results
        num_y = len(y)
        resp = {}

        if num_y != num_outputs:
            return "expected %d output classes, got inference with %d results" % (num_outputs, num_y), 500

        resp = { classes[i] : y[i] for i in range(num_y) }

        return jsonify(resp), 200
    except Exception as e:
        return str(e), 400

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo serve", description="Load a model and expose an API that can be used to run its inference.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path", help="Path of the project.")

    parser.add_argument("-a", "--address", dest="address", action="store", type=str,
        help="IP address or hostname to bind to.")
    parser.add_argument("-p", "--port", dest="port", action="store", type=int, default=8080,
        help="TCP port to bind to.")
    parser.add_argument("--debug", dest="debug", action="store_true", default=False,
        help="Enable debug messages.")

    parser.add_argument("--profile", dest="profile", action="store_true", default=False,
        help="Enable profiler.")
    parser.add_argument("--prof-max", dest="restrictions", type=int, default=30,
        help="Maximum number of calls to profile.")

    parser.add_argument("--classes", dest="classes", default=None,
        help="Optional comma separated list of output classes.")

    args = parser.parse_args(argv)
    return args

def action_serve(argc, argv):
    global prj, app, classes, num_outputs

    args = parse_args(argv)
    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    elif not prj.is_trained():
        log.error("no trained Keras model found for this project")
        quit()

    if args.classes is None:
        num_outputs = prj.model.output.shape[1]
        if prj.classes is None:
            classes = ["class_%d" % i for i in range(num_outputs)]
        else:
            classes = [prj.classes[i] for i in range(num_outputs)]
    else:
        classes = [s.strip() for s in args.classes.split(',') if s.strip() != ""]
        num_outputs = len(classes)

    if args.profile:
        from werkzeug.contrib.profiler import ProfilerMiddleware
        args.debug = True
        app.config['PROFILE'] = True
        app.wsgi_app = ProfilerMiddleware(app.wsgi_app, restrictions=[args.restrictions])

    app.run(host=args.address, port=args.port, debug=args.debug)



import os
import argparse
import logging as log

from ergo.project import Project

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo view", description="View the model struture and training statistics.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("path", help="Path of the project.")
    parser.add_argument( "--img-only", dest="img_only", default=False, action="store_true",
        help="Save plots as PNG files but don't show them in a UI.")
    args = parser.parse_args(argv)
    return args

def action_view(argc, argv):
    args = parse_args(argv)
    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()

    prj.view(args.img_only)



class Template(object):
    def __init__(self, name, code):
        self.name = name
        self.code = code.strip()

    def compile(self, ctx):
        compiled = self.code
        for key, val in ctx.items():
            compiled = compiled.replace("{%s}" % key, str(val))
        return compiled



import os
import json
import itertools
import numpy as np
import logging as log

def model(prj, img_only):
    if prj.model is not None:
        prj.model.summary()

def roc(prj, img_only):
    if prj.dataset.has_test():
        import matplotlib.pyplot as plt
        from sklearn.metrics import roc_curve, auc

        log.info("found %s, loading ...", prj.dataset.test_path)
        prj.dataset.load_test()
        log.info("computing ROC curve on %d samples ...", len(prj.dataset.X_test))

        y_pred = prj.model.predict(prj.dataset.X_test)
        fpr, tpr, thresholds = roc_curve(prj.dataset.Y_test.ravel(), y_pred.ravel())

        plt.figure("ROC Curve")
        plt.title("ROC Curve")
        plt.plot([0, 1], [0, 1], 'k--')
        plt.plot(fpr, tpr, label='AUC = {:.3f}'.format(auc(fpr, tpr)))
        plt.xlabel('FPR')
        plt.ylabel('TPR')
        plt.legend()

        plt.savefig( os.path.join(prj.path, 'roc.png') )

def stats(prj, img_only):
    if os.path.exists(prj.txt_stats_path):
        with open(prj.txt_stats_path, 'rt') as fp:
            print(fp.read().strip())

    if os.path.exists(prj.json_stats_path):
        import matplotlib.pyplot as plt

        with open(prj.json_stats_path, 'rt') as fp:
            stats = json.load(fp)
            for who, header in prj.what.items():
                orig = np.array(stats[who]['cm'])
                cm = np.array(stats[who]['cm'])
                tot = cm.sum()
                cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

                name = header.strip(" -\n").lower()
                title = "%s confusion matrix (%d samples)" % (name, tot)
                filename = os.path.join(prj.path, "%s_cm.png" % name)

                plt.figure(title)
                plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Reds)
                plt.title(title)
                plt.colorbar()
                classes = range(0, cm.shape[0])
                tick_marks = np.arange(len(classes))
                plt.xticks(tick_marks, classes, rotation=45)
                plt.yticks(tick_marks, classes)
                thresh = cm.max() / 2.
                for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                    plt.text(j, i, "%.1f%% (%d)" % (cm[i, j] * 100, orig[i, j]),
                             horizontalalignment="center",
                             color="white" if cm[i, j] > thresh else "black")

                plt.tight_layout()
                plt.ylabel('truth')
                plt.xlabel('prediction')
                plt.savefig(filename)

def history(prj, img_only):
    if prj.history is not None:
        import matplotlib.pyplot as plt

        plt.figure("training history")
        # Plot training & validation accuracy values
        plt.subplot(2,1,1)
        plt.plot(prj.history['acc'])
        plt.plot(prj.history['val_acc'])
        plt.title('Model accuracy')
        plt.ylabel('Accuracy')
        plt.xlabel('Epoch')
        plt.legend(['Train', 'Test'], loc='lower right')

        # Plot training & validation loss values
        plt.subplot(2,1,2)
        plt.plot(prj.history['loss'])
        plt.plot(prj.history['val_loss'])
        plt.title('Model loss')
        plt.ylabel('Loss')
        plt.xlabel('Epoch')
        plt.legend(['Train', 'Test'], loc='upper right')
        plt.tight_layout()
        plt.savefig( os.path.join(prj.path, 'history.png') )

def plot_clusters(prj, pca, X, y, ca, D3 = False):
    import matplotlib.pyplot as plt
    from numpy import argmax
    from matplotlib.lines import Line2D
    from mpl_toolkits.mplot3d import axes3d

    Xt = pca.transform(X)
    Xt = Xt[:, :3]
    y = argmax(y, axis=1)

    clusters = ca.labels_
    cmap = plt.cm.jet
    cmaplist = [cmap(i) for i in range(0, cmap.N, cmap.N//len(set(clusters)))]
    cmap = cmap.from_list('Custom cmap', cmaplist, cmap.N)

    if D3:
        fig = plt.figure('Clustering Analisys - 3D projection')
        ax = fig.add_subplot(1, 1, 1, projection = '3d')
        ax.set_zlabel('Principal component 3')
    else:
        fig = plt.figure('Clustering Analisys')
        ax = fig.add_subplot(1, 1, 1)

    ax.set_xlabel('Principal component 1')
    ax.set_ylabel('Principal component 2')
    markers = [m for m, func in Line2D.markers.items()
                        if func != 'nothing' and m in Line2D.filled_markers]
    _legends = []
    for cl in list(set(y)):
        idx =  np.where(y == cl)[0]
        if D3:
            scatter = ax.scatter(Xt[idx, 0], Xt[idx, 1], Xt[idx,2], cmap=cmap, s=40, alpha=0.5,
                                 c=clusters[idx],
                                 marker=markers[cl],
                                 label=y[idx],
                                 edgecolor='black'
                                 )
        else:
            scatter = ax.scatter(Xt[idx,0], Xt[idx,1], cmap = cmap, s = 40, alpha = 0.5,
                                 c = clusters[idx],
                                 marker = markers[cl],
                                 label=y[idx],
                                 edgecolor='black'
                                )
        l = Line2D([], [], color='black', marker=markers[cl], linestyle='None',
                          markersize=5, mfc = 'w', markeredgecolor = 'black', label='Class %d' % cl)
        _legends.append(l)

    for cl in list(set(clusters)):
        label = 'Cluster %d' % cl
        if cl == -1:
            label = 'Noisy samples'
        l = Line2D([], [], color= cmaplist[cl], marker='o', linestyle='None',
                          markersize=5,  label=label)
        _legends.append(l)

    legend1 = ax.legend(handles=_legends, loc='upper left')
    ax.add_artist(legend1)

    fig.tight_layout()
    if D3:
        name = 'clusters_projection_3D.png'
    else:
        name = 'clusters_projection.png'
    fig.savefig(os.path.join(prj.path, name))

def plot_intertia(prj, x, y):
    import matplotlib.pyplot as plt

    fig = plt.figure("Inertia of Kmeans clustering algorithm")
    ax = fig.add_subplot(1,1,1)
    ax.set_xlabel('# of clusters')
    ax.set_ylabel('Intertia')
    ax.plot(x,y, '*-')
    fig.savefig(os.path.join(prj.path, 'kmeans_inertia.png'))


def pca_projection(prj, pca, X, y, D3):
    import matplotlib.pyplot as plt
    from numpy import argmax
    from mpl_toolkits.mplot3d import axes3d

    Xt = pca.transform(X)
    Xt = Xt[:,:3]
    y = argmax(y, axis=1)

    cmap = plt.cm.jet
    cmaplist = [cmap(i) for i in range(cmap.N)]
    cmap = cmap.from_list('Custom cmap', cmaplist, cmap.N)

    if D3:
        fig = plt.figure('PCA decomposition - 3D projection')
        ax = fig.add_subplot(1, 1, 1, projection = '3d')
        ax.set_zlabel('Principal component 3')
    else:
        fig = plt.figure('PCA decomposition')
        ax = fig.add_subplot(1, 1, 1)

    ax.set_xlabel('Principal component 1')
    ax.set_ylabel('Principal component 2')

    if D3:
        scatter = ax.scatter(Xt[:, 0], Xt[:, 1], Xt[:,2], c=y, cmap=cmap, label = y, s=5, alpha=0.5)
    else:
        scatter = ax.scatter(Xt[:, 0], Xt[:, 1], c=y, cmap=cmap, label=y, s=5, alpha=0.5)
    legend1 = ax.legend(*scatter.legend_elements(), loc = 'upper right', title = 'Class')
    ax.add_artist(legend1)
    fig.tight_layout()
    if D3:
        name = 'pca_projection_3d.png'
    else:
        name = 'pca_projection.png'
    fig.savefig( os.path.join(prj.path, name))


def pca_explained_variance(prj, pca, img_only):
    import matplotlib.pyplot as plt
    exp = pca.explained_variance_ratio_.cumsum()

    exp90, exp95, exp99 = -1, -1, -1
    for i,j in enumerate(exp):
        if j >= 0.9 and exp90 == -1:
            exp90 = i
        elif j >= 0.95 and exp95 == -1:
            exp95 = i
        elif j >= 0.99 and exp99 == -1:
            exp99 = i

    fig = plt.figure('PCA explained variance')
    ax = fig.add_subplot(1,1,1)
    ax.set_xlabel('Principal component number')
    ax.set_ylabel('Explained variance')
    ax.plot(exp, '-')

    # show 90, 95 and 99 % explanation
    ax.axvline(x=exp90, label='%d PC > 90%%' % exp90, linestyle = '--', c='k')
    ax.axvline(x=exp95, label='%d PC > 95%%' % exp95, linestyle = '--', c='b')
    ax.axvline(x=exp99, label='%d PC > 99%%' % exp99, linestyle = '--', c='r')
    ax.legend(title = 'Required components')

    fig.tight_layout()
    fig.savefig(os.path.join(prj.path, 'pca_explained_ratio.png'))


def correlation_matrix(prj, corr, img_only):
    import numpy as np
    import seaborn as sns
    import matplotlib.pyplot as plt

    sns.set(style="white")
    mask = np.zeros_like(corr, dtype=np.bool)
    mask[np.triu_indices_from(mask)] = True

    fig, ax = plt.subplots()
    cmap = sns.diverging_palette(220, 10, as_cmap=True)
    sns.heatmap(corr, square = True, mask=mask, cmap=cmap,cbar_kws={"shrink": .7})
    fig.savefig(os.path.join(prj.path, 'corr_matrix.png'))

def show(img_only):
    if not img_only:
        import matplotlib.pyplot as plt
        plt.show()




import os
import argparse
import logging as log

from ergo.core.optimizer import optimize_dataset
from ergo.dataset import Dataset

def probability(x):
    x = float(x)
    if x < 0 or x > 1:
        raise argparse.ArgumentTypeError("%r not in range [0.0 - 1.0]" % x )
    return x

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo optimize-dataset", description="Remove duplicates from the dataset and only allow them within a given reuse ratio.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path", help="Path of the project to create.")

    parser.add_argument("-r", "--reuse-ratio", dest="reuse", action="store", type=probability, default=0.15,
        help="Reuse ratio of duplicates in the [0,1] range.")
    parser.add_argument("-o", "--output", dest="output", action="store", default=None,
        help="Output dataset file.")

    args = parser.parse_args(argv)
    return args

def action_optimize_dataset(argc, argv):
    args = parse_args(argv)
    path = os.path.abspath(args.path)
    if not os.path.exists(path):
        log.error("dataset file %s does not exist", path)
        quit()
    
    optimize_dataset(path, args.reuse, args.output)



import argparse
import os
import logging as log

from ergo.project import Project
from ergo.core.utils import clean_if_exist

def probability(x):
    x = float(x)
    if x < 0 or x > 1:
        raise argparse.ArgumentTypeError("%r not in range [0.0 - 1.0]" % x)
    return x


def validate_args(args):
    if args.dataset is not None and (not args.dataset.startswith('sum://') and not os.path.exists(args.dataset)):
        log.error("file %s does not exist" % args.dataset)
        quit()

    p_train = 1 - args.test - args.validation
    if p_train < 0:
        log.error("validation and test proportion are bigger than 1")
        quit()
    elif p_train < 0.5:
        log.error("using less than 50% of the dataset for training")
        quit()

def clean_dataset(path):
    files = [ \
        'data-train.csv',
        'data-test.csv',
        'data-validation.csv',
        'data-train.pkl',
        'data-test.pkl',
        'data-validation.pkl']

    clean_if_exist(path, files)

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo prepare", description="Create train, test and validation sets.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("path", help="The path containing the model definition.")

    parser.add_argument("-d", "--dataset", action="store", dest="dataset",
                        help="Path of the dataset to use or a SUM server address. Leave empty to reuse previously generated subsets.")
    parser.add_argument("-t", "--test", action="store", dest="test", type=probability, default=0.15,
                        help="Proportion of the test set in the (0,1] interval.")
    parser.add_argument("-v", "--validation", action="store", dest="validation", type=probability, default=0.15,
                        help="Proportion of the validation set in the (0,1] interval.")
    parser.add_argument("--no-shuffle", dest="no_shuffle", action="store_true", default=False,
                        help="Do not shuffle dataset during preparation.")

    args = parser.parse_args(argv)
    validate_args(args)
    return args


def action_prepare(argc, argv):
    args = parse_args(argv)
    prj = Project(args.path)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()

    if args.dataset is None:
        log.error("no --dataset argument specified", args.path)
        quit()

    if prj.dataset.exists():
        log.info("removing previously generated datasets")
        clean_dataset(args.path)

    prj.dataset.do_save = True
    prj.prepare(args.dataset, args.test, args.validation, not args.no_shuffle)


import sys
import os
import argparse
import logging as log

from ergo.project import Project
from ergo.templates import Templates

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo create", description="Create a new ergo project.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("path", help="Path of the project to create.")
    
    parser.add_argument("-i", "--inputs", dest="num_inputs", action="store", type=int, default=10,
        help="Number of inputs of the model.")
    parser.add_argument("-o", "--outputs", dest="num_outputs", action="store", type=int, default=2,
        help="Number of outputs of the model.")
    parser.add_argument("-l", "--layers", dest="hidden", action="store", type=str, default="30, 30",
        help="Comma separated list of positive integers, one per each hidden layer representing its size.")
    parser.add_argument("-b", "--batch-size", dest="batch_size", action="store", type=int, default=64,
        help="Batch size parameter for training.")
    parser.add_argument("-e", "--epochs", dest="max_epochs", action="store", type=int, default=50,
        help="Maximum number of epochs to train the model.")

    args = parser.parse_args(argv)
    return args

def action_create(argc, argv):
    args = parse_args(argv)
    if os.path.exists(args.path):
        log.error("path %s already exists" % args.path)
        quit()

    check = [n for n in [int(s.strip()) for s in args.hidden.split(',') if s.strip() != ""] if n > 0]
    if len(check) < 1:
        log.error("the --hidden argument must be a comma separated list of at least one positive integer")
        quit()

    ctx = {
        'NUM_INPUTS': args.num_inputs,
        'HIDDEN':     ', '.join([str(n) for n in check]),
        'NUM_OUTPUTS': args.num_outputs,
        'BATCH_SIZE': args.batch_size,
        'MAX_EPOCHS': args.max_epochs,
    }

    log.info("initializing project %s with ANN %d(%s)%d ...", args.path, ctx['NUM_INPUTS'], ctx['HIDDEN'], ctx['NUM_OUTPUTS'])
    os.makedirs(args.path, exist_ok=True)
    for tpl in Templates:
        log.info( "creating %s", tpl.name)
        with open( os.path.join(args.path, tpl.name), 'wt' ) as fp:
            data = tpl.compile(ctx)  
            fp.write(data)



import os
import time
import sys
import glob
import argparse
import logging as log
import multiprocessing

from ergo.project import Project
from ergo.core.queue import TaskQueue

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ergo encode", description="Encode one or more files to vectors and create or update a csv dataset for training.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("project", help="The path containing the model definition.")
    parser.add_argument("path", help="Path of a single file or of a folder of files.")

    parser.add_argument( "-l", "--label", dest="label", default='auto',
        help="The class of the file(s) or 'auto' to use the name of the containing folder.")
    parser.add_argument( "-o", "--output", dest="output", default='dataset.csv',
        help="Output CSV file names.")
    parser.add_argument( "-f", "--filter", dest="filter", default='*.*',
        help="If PATH is a folder, this flag determines which files are going to be selected.")
    parser.add_argument( "-m", "--multi", dest="multi", default=False, action="store_true",
        help="If PATH is a single file, this flag will enable line by line reading of multiple inputs from it.")
    parser.add_argument( "-w", "--workers", dest="workers", default=0, type=int,
        help="Number of concurrent workers to use for encoding, or zero to run two workers per available CPU core.")
    parser.add_argument( "-d", "--delete", dest="delete", default=False, action="store_true",
        help="Delete each file after encoding it.")

    args = parser.parse_args(argv)
    return args

# given the arguments setup and a path, return the label
def label_of(args, path):
    if args.label == 'auto':
        return os.path.basename(os.path.dirname(path))
    else:
        return args.label

progress_at = None
prev_done = 0
prev_speed = 0

def get_speed(done):
    global progress_at, prev_done, prev_speed

    speed = 0
    now   = time.time()
    if progress_at is not None:
        delta_v = int(done - prev_done)
        if delta_v >= 10:
            delta_t = now - progress_at
            speed = delta_v / delta_t
        else:
            speed = prev_speed
            now   = progress_at
            done  = prev_done

    progress_at = now
    prev_done = done
    prev_speed = speed

    return speed

# simple progress bar
def on_progress(done, total):
    maxsz = 80
    perc  = done / total
    sp    = " " if perc > 0 else ""
    bar   = ("█" * int(maxsz * perc)) + sp

    sys.stdout.write("\r%d/%d (%d/s) %s%.1f%%" % (done, total, get_speed(done), bar, perc * 100.0))
    if perc == 1.0:
        sys.stdout.write("\n")

# wait for lines on the results queue and append them to filepath
def appender(filepath, total, results):
    log.debug("file appender for %s started ...", filepath)

    print()
    with open(filepath, 'a+t') as fp:
        done = 0
        while True: 
            on_progress(done, total)

            line = results.get()
            if line is None:
                break

            fp.write(line.strip() + "\n")
            done += 1

# use the project prepare_input function to encode a single
# input into a CSV line
def parse_input(prj, inp, label, results, do_delete = False):
    log.debug("encoding '%s' as '%s' ...", inp, label)
    x = prj.logic.prepare_input(inp, is_encoding = True)
    results.put("%s,%s" % (str(label), ','.join(map(str,x))))
    if do_delete and os.path.isfile(inp): 
        os.remove(inp)

def action_encode(argc, argv):
    args = parse_args(argv)
    if not os.path.exists(args.path):
        log.error("%s does not exist.", args.path)
        quit()

    prj = Project(args.project)
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    
    args.label = args.label.strip().lower()
    log.info("using %s labeling", 'auto' if args.label == 'auto' else 'hardcoded')

    inputs = []
    if os.path.isdir(args.path):
        in_files = []
        if args.label == 'auto':
            # the label is inferred from the dirname, so we expect
            # args.path to contain multiple subfolders
            for subfolder in glob.glob(os.path.join(args.path, "*")):
                log.info("enumerating %s ...", subfolder)
                in_filter = os.path.join(subfolder, args.filter)
                in_sub    = glob.glob(in_filter)
                n_sub     = len(in_sub)
                if n_sub > 0:
                    log.info("collected %d inputs from %s", n_sub, subfolder)
                    in_files.extend(in_sub)
        else:
            # grab files directly from args.path
            in_filter = os.path.join(args.path, args.filter)
            in_files.extend(glob.glob(in_filter))
            log.info("collected %d inputs from %s", len(in_files), args.path)

        log.info("labeling %d files ...", len(in_files))
        for filepath in in_files:
            if os.path.isfile(filepath):
                inputs.append((label_of(args, filepath), filepath))

    elif args.multi:
        log.info("parsing multiple inputs from %s ...", args.path)
        label = label_of(args, args.path)
        with open(args.path, 'rt') as fp:
            for line in fp:
                inputs.append((label, line))

    else:
        label = label_of(args, args.path)
        inputs.append((label, args.path))

    # one encoding queue that pushes to another queue that centralizes
    # append operations to a single writer process
    num_in = len(inputs)
    enc_q = TaskQueue('encoding', args.workers)
    res_q = multiprocessing.Queue()
    app_p = multiprocessing.Process(target=appender, args=(args.output, num_in, res_q))

    # open the output file and start waiting for lines to append
    app_p.start()

    log.info("encoding %d inputs to %s ...", num_in, args.output)
    for (y, x) in inputs:
        enc_q.add_task(parse_input, prj, x, y, res_q, args.delete)

    # wait for all inputs to be encoded
    enc_q.join()
    # let the writer know there are no more inputs to read
    res_q.put(None)
    # wait for the writer to finish
    app_p.join()




import os
import shutil
import logging as log
import numpy as np

from collections import defaultdict

def clean_if_exist(path, files):
    path  = os.path.abspath(path)
    for filename in files:
        filename = os.path.join(path, filename)
        if os.path.exists(filename):
            if os.path.isdir(filename):
                log.info("removing folder %s", filename)
                shutil.rmtree(filename)
            else:
                log.info("removing file %s", filename)
                os.remove(filename)

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def serialize_classification_report(cr):
    tmp = list()
    for row in cr.split("\n"):
        parsed_row = [x.strip() for x in row.split("  ") if len(x.strip()) > 0]
        if len(parsed_row) > 0:
            tmp.append(parsed_row)
    
    measures = tmp[0]
    out = defaultdict(dict)
    for row in tmp[1:]:
        columns      = len(row)
        class_label  = row[0].strip()
        num_measures = len(measures)
        
        # fixes https://github.com/evilsocket/ergo/issues/5
        while columns < num_measures + 1:
            row.insert(1, None)
            columns += 1

        for j, m in enumerate(measures):
            v = row[j + 1]
            value = float(v.strip()) if v is not None else None
            metric = m.strip()
            out[class_label][metric] = value

    return out

def serialize_cm(cm):
    return cm.tolist()



from threading import Thread
import multiprocessing
import logging as log
import queue 
import time
import os

class TaskQueue(queue.Queue):
    def __init__(self, name, num_workers=-1, blocking=False):
        queue.Queue.__init__(self)
        self.name = name
        self.blocking = blocking
        self.num_workers = num_workers if num_workers > 0 else (multiprocessing.cpu_count() * 2)
        self._start_workers()

    def add_task(self, task, *args, **kwargs):
        args = args or ()
        kwargs = kwargs or {}
        self.put((task, args, kwargs), block=self.blocking)

    def _start_workers(self):
        log.info("starting %d workers for %s" % (self.num_workers, self.name))
        for i in range(self.num_workers):
            t = Thread(target=self._worker)
            t.daemon = True
            t.start()

    def _worker(self):
        log.debug("task queue worker started")
        while True:
            item, args, kwargs = self.get()
            item(*args, **kwargs)  
            self.task_done()

